# -*- encoding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ironic.api.middleware import auth_token
from ironic.api.middleware import json_ext
from ironic.api.middleware import parsable_error


ParsableErrorMiddleware = parsable_error.ParsableErrorMiddleware
AuthTokenMiddleware = auth_token.AuthTokenMiddleware
JsonExtensionMiddleware = json_ext.JsonExtensionMiddleware

__all__ = ('ParsableErrorMiddleware',
           'AuthTokenMiddleware',
           'JsonExtensionMiddleware')
