# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add resource_class to node

Revision ID: dd34e1f1303b
Revises: 10b163d4481e
Create Date: 2016-07-20 21:48:12.475320

"""

# revision identifiers, used by Alembic.
revision = 'dd34e1f1303b'
down_revision = '10b163d4481e'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('nodes', sa.Column('resource_class', sa.String(80),
                                     nullable=True))
